/*--------------------------------------------------------------------
 * $Id: ipa.c,v 1.5 2001/06/20 14:30:58 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/wcloud3d.h"


typedef struct {
  int ix;       /* x-index of the pixel     */
  int iy;       /* y-index of the pixel     */
  int gspec;    /* g specified ?            */
  int wspec;    /* w specified ?            */

  int nc;       /* number of cloudy pixels  */
  double tau;   /* total optical depth      */
 
  float ***k;  /* extinction coefficient   */
  float ***g;  /* asymmetry parameter      */
  float ***w;  /* single-scattering albedo */

  int nz;       /* number of layers         */
}  sortarray_struct;


#define EPSILON 1E-6


/* prototypes of internal functions */
static int compare_levels (double *z1, int n1,
			   float *z2, int n2);
static int sortarray_fnc (const void *x1, const void *x2);
static int write_pixel (double *z3D, float ***k3D, float ***g3D, float ***w3D, 
			int *threed, int gspec, int wspec, 
			int ix, int iy, int Nz, int index,
			output_struct *output);
static int threeDtoIPA (double *z3D, float ***k3D, float ***g3D, float ***w3d, 
			int *threed, int gspec, int wspec, int Nx, int Ny, int Nz,
			output_struct *output);
static int read_wc_ipa (input_struct input, output_struct *output);
static int read_ic_ipa (input_struct input, output_struct *output);



int setup_ipa (input_struct input, output_struct *output)
{
  int status=0;

  int ix=0, ip=0, lev=0;
  int Nx=0, Ny=0, Nz=0;
  double delX=0, delY=0;
  double *z3D=NULL;
  float ***k3D=NULL, ***g3D=NULL, ***w3D=NULL;
  int *threed=NULL;
  int gspec=0, wspec=0;

  char wcloud3Dfilename[FILENAME_MAX]="wcloud3D.dat";

  /* we need a least one independent pixel */
  output->nipa = 1;


  if (strlen(input.filename[FN_WCLOUD_IPA])>0 || 
      strlen(input.filename[FN_ICLOUD_IPA])>0) {
    
    /* read water cloud IPA file */
    if (strlen(input.filename[FN_WCLOUD_IPA])>0) {
      status = read_wc_ipa (input, output);
      if (status!=0) {
	fprintf (stderr, "Error %d returned by read_wc_ipa()\n", status);
	return status;
      }
    }

    /* read ice cloud IPA file; this function contains a test */
    /* if the weights defined in the ice cloud file are       */
    /* identical to those in the water cloud file             */

    if (strlen(input.filename[FN_ICLOUD_IPA])>0) {
      status = read_ic_ipa (input, output);
      if (status!=0) {
	fprintf (stderr, "Error %d returned by read_ic_ipa()\n", status);
	return status;
      }
    }
  }
  else  {

    if (input.wcipa)  {
      if (input.cloudcover < 0) {  /* IPA, clouds read from wcloud3D.dat */
	

	/* read 3D cloud data */
	status = read_3D_cloud (wcloud3Dfilename,
				&Nx, &Ny, &Nz,
				&delX, &delY,
				&z3D, &k3D, &threed,
				&g3D, &gspec,
				&w3D, &wspec,
				input.quiet);
	
	if (status!=0) {
	  fprintf (stderr, "error %d reading %s\n", status, wcloud3Dfilename);
	  return status;
	}
	
	
	/******************************/
	/* compare 3D and 1D profiles */
	/******************************/
	
	/* make sure that 1D and 3D properties are defined on the same altitude grid */
	status = compare_levels (z3D, Nz, output->atm.zd, output->atm.nlyr);
	
	if (status!=0) {
	  fprintf (stderr, "\nFATAL error: z-grids of 1D atmosphere and 3D clouds are different\n");
	  return status;
	}
	
	if (!input.quiet)
	  fprintf (stderr, " ... 3D cloud definition seems to be o.k.\n");
	
	
	/* allocate memory for IPA arrays */
	output->wcipa     = calloc (Nx*Ny, sizeof(cld_out_struct));
	output->ipaweight = calloc (Nx*Ny, sizeof(float));
	
	status = threeDtoIPA (z3D, k3D, g3D, w3D, threed, gspec, wspec,
			      Nx, Ny, Nz,
			      output);
	
	if (status!=0) {
	  fprintf (stderr, "error %d creating IPA input\n", status);
	  return status;
	}
	

	if (!input.quiet)
	  fprintf (stderr, " ... CONGRATULATIONS! You only need to consider %d independent pixels!\n", 
		   output->nipa);
	
	/*************************************************************/
	/* output->nipa independent pixels are now stored in         */
	/* output->wcipa[ip] where ip = 0 ... ntau is the index of   */
	/* the pixel and output->weight[ip] is the correspoding      */
	/* weight.                                                   */
	/*************************************************************/
      
	/* free memory */
	for (lev=0; lev<Nz; lev++)
	  for (ix=0; ix<Nx; ix++)
	    free(k3D[lev][ix]);
	free (k3D);
	
	if (gspec) {
	  for (lev=0; lev<Nz; lev++)
	    for (ix=0; ix<Nx; ix++)
	      free(g3D[lev][ix]);
	  free (g3D);
	}
	
	if (wspec) {
	  for (lev=0; lev<Nz; lev++)
	    for (ix=0; ix<Nx; ix++)
	      free(w3D[lev][ix]);
	  free (w3D);
	}
	
	free(threed);
	free(z3D);
      
	return 0;
      }
      else  {  /* cloud cover defined in input file */
	
	if (!input.quiet)
	  fprintf (stderr, " ... IPA with two columns, cloudless and overcast\n");
	
	/* allocate memory for IPA arrays */
	output->wcipa     = calloc (2, sizeof(cld_out_struct));
	output->ipaweight = calloc (2, sizeof(float));
	
	for (ip=0; ip<2; ip++) {
	  
	  status = calloc_cld_out (&(output->wcipa[ip]), 
				   output->wl.nlambda_t, 
				   output->atm.nlev);

	  if (status!=0) {
	    fprintf (stderr, "Error %d allocating memory for cld_out_struct\n", status);
	    return status;
	  }
	}      
	
	output->nipa=2;                              /* 2 independent pixels */
	
	output->ipaweight[0] = 1.0-input.cloudcover; /* cloudless pixel      */
	output->ipaweight[1] = input.cloudcover;     /* overcast pixel       */
	
	/* copy 1D cloud to 2nd pixel, first pixel is left as it is (cloudless) */
	status = cp_cld_out (&(output->wcipa[1]), output->wc, 0);
	if (status!=0) {
	  fprintf (stderr, "Error %d copying output->wc to output->wcipa[1]\n", status);
	  return status;
	}
      }
    }
    else {  /* no independent pixel */
      output->nipa=1;
      output->ipaweight=calloc (1, sizeof(double));
      output->ipaweight[0] = 1;
    }
  }

  return 0;   /* if o.k. */
}




/*********************************/
/* compare two altitude grids    */
/*********************************/

static int compare_levels (double *z1, int n1,
			   float *z2, int n2)
{
  int kc=0;

  if (n1!=n2) {
    fprintf (stderr, "Error, number of layers differing\n");
    fprintf (stderr, "%d in 3D, %d in 1D\n", n1, n2);

    return -1;
  }

  for (kc=0; kc<=n1; kc++)
    if (fabs((float) (z1[kc]/1000.0) - z2[n1-kc]) > EPSILON * z1[kc]) {
      fprintf (stderr, "Error, difference at level %d, altitude %g vs. %g\n",	
	       kc, z1[kc], z2[n1-kc]);
      return -1;
    }

  return 0;  /* if o.k. */
}


/***********************************************************************************/
/* Function: sortarray_fnc                                                         */
/* Description: Internal function, required by ASCII_sortarray();                  */
/*              see description of standard C library function qsort().            */
/*                                                                                 */
/* Parameters:                                                                     */    
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/***********************************************************************************/

static int sortarray_fnc (const void *x1, const void *x2)
{
  sortarray_struct *c1 = (sortarray_struct *) x1;
  sortarray_struct *c2 = (sortarray_struct *) x2;

  int lev=0;
  double sub;

  if (c1->tau > c2->tau)
    return 1;
  
  if (c1->tau < c2->tau)
    return -1;

  /*****************************************************************/
  /* equal optical depths, need to compare number of cloudy pixels */
  /*****************************************************************/
  
  if (c1->nc > c2->nc)
    return 1;
  
  if (c1->nc < c2->nc)
    return -1;

  /******************************************************************************/
  /* equal optical depths and number of cloudy pixels, need to compare profiles */
  /******************************************************************************/

  sub=0;
  for (lev=0; lev<c1->nz; lev++) {
    
    /* compare extinction coefficients */
    if ((sub = c1->k[lev][c1->ix][c1->iy] - c2->k[lev][c2->ix][c2->iy]) != 0)
      break;
  
    /* compare asymmetry parameters */
    if (c1->gspec)
      if ((sub = c1->g[lev][c1->ix][c1->iy] - c2->g[lev][c2->ix][c2->iy]) != 0)
	break;
  
    /* compare single scattering albedo */
    if (c1->wspec)
      if ((sub = c1->w[lev][c1->ix][c1->iy] - c2->w[lev][c2->ix][c2->iy]) != 0)
	break;
  }  

  if (lev==c1->nz)   /* if reached TOA without finding different layers */
    return 0;
  
  return (sub>0?-1:1);
}





/**********************************************************/
/* Write one column of a cloud, used by threeDtoIPA().    */ 
/**********************************************************/

static int write_pixel (double *z3D, float ***k3D, float ***g3D, float ***w3D, 
			int *threed, int gspec, int wspec, 
			int ix, int iy, int Nz, int index,
			output_struct *output)
{
  int iv=0, lev=0, status=0;

  /* allocate memory for this pixel; allocate memory for */
  /* ALL fields even if they are not required because    */
  /* they may be used by other functions.                */

  status = calloc_cld_out (&(output->wcipa[index]), output->wl.nlambda_t, output->atm.nlev);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cld_out_struct\n", status);
    return status;
  }

  /* copy 1D cloud (from wc_file) to wcipa[index] */
  status = cp_cld_out (&(output->wcipa[index]), output->wc, 0);
  if (status!=0) {
    fprintf (stderr, "Error %d copying output->wc to output->wcipa[index]\n", status);
    return status;
  }
  

  /* profiles need to be reversed ! */
  for (lev=0; lev<Nz; lev++)
    for (iv=0; iv<output->wl.nlambda_t; iv++) {

      /* overwrite only levels where actually a 3D cloud is specified */
      if (threed[lev]) {
	
	output->wcipa[index].optprop.dtau[iv][Nz-lev-1] = k3D[lev][ix][iy]*(z3D[lev+1]-z3D[lev]);
	
	if (gspec)
	  output->wcipa[index].optprop.gg[iv][Nz-lev-1]  = g3D[lev][ix][iy];
	
	if (wspec)
	  output->wcipa[index].optprop.ssa[iv][Nz-lev-1] = w3D[lev][ix][iy];
      }
    }
  
  return 0;
}





/*****************************************/
/* Write 1D cloud data                   */
/*****************************************/

static int threeDtoIPA (double *z3D, float ***k3D, float ***g3D, float ***w3D, 
			int *threed, int gspec, int wspec, int Nx, int Ny, int Nz,
			output_struct *output)
{
  int keyc=0, index=0;
  int ix=0, iy=0, lev=0, status=0;

  FILE *cldfile=NULL;
  char cldfilename    [FILENAME_MAX+1] = "ipa.idx";
  char cldcolfilename [FILENAME_MAX+1] = "";

  int sarrayc=0;
  int narray = Nx*Ny;

  sortarray_struct *sarray  = calloc(narray, sizeof(sortarray_struct));
  sortarray_struct *sarray1 = calloc(1, sizeof(sortarray_struct));

  fprintf (stderr, " ... identifying identical columns for IPA\n");
  
  sarrayc=0;
  for (ix=0; ix<Nx; ix++)
    for (iy=0; iy<Ny; iy++) {
      sarray[sarrayc].nz = Nz;
      sarray[sarrayc].ix = ix;
      sarray[sarrayc].iy = iy;
      sarray[sarrayc].k  = k3D;
      sarray[sarrayc].g  = g3D;
      sarray[sarrayc].w  = w3D;

      sarray[sarrayc].gspec = gspec;
      sarray[sarrayc].wspec = wspec;

      /* total vertical optical depth */
      sarray[sarrayc].tau   = 0;
      for (lev=0; lev<Nz; lev++)
	sarray[sarrayc].tau += k3D[lev][ix][iy] * (z3D[lev+1] - z3D[lev]);

      /* cloudy pixel counter */
      sarray[sarrayc].nc=0;
      for (lev=0; lev<Nz; lev++)
	if (k3D[lev][ix][iy]>0)
	  sarray[sarrayc].nc++;

      sarrayc++;
    }
  
  qsort (sarray, narray, sizeof(sortarray_struct), sortarray_fnc);
  
  /* create filename for index file */
  cldfile = fopen (cldfilename, "w");
  fprintf (cldfile, "# Cloud optical depth distribution, generated by %s\n", "uvspec");
  fprintf (cldfile, "# File number, filename, number of cloudy pixels, optical depth, number of occurrences\n");


  sarrayc = 0;
  memcpy (sarray1, &(sarray[0]), sizeof(sortarray_struct)); 
  keyc = 0;
  index=0;

  for (sarrayc=0; sarrayc<narray; sarrayc++) {
    
    /* if different column */
    if (sortarray_fnc (sarray1, &(sarray[sarrayc])) != 0) {
      
      /* write one pixel */
      status = write_pixel (z3D, k3D, g3D, w3D, threed, gspec, wspec,
			    (sarray[sarrayc-1]).ix, (sarray[sarrayc-1]).iy, 
			    Nz, index, output);

      /* entry in index file */

      fprintf (cldfile, "%d %s %d %f %d\n", 
	       index, cldcolfilename, 
	       (sarray[sarrayc-1]).nc, 
	       (sarray[sarrayc-1]).tau,
	       keyc);
      
      output->ipaweight[index] = (double) keyc / (double) narray;

      memcpy (sarray1, &(sarray[sarrayc]), sizeof(sortarray_struct)); 
      keyc=1;
      index++;
    }
    else 
      keyc++;
  }
  
  /* write last pixel */
  status = write_pixel (z3D, k3D, g3D, w3D, threed, gspec, wspec,
			(sarray[sarrayc-1]).ix, (sarray[sarrayc-1]).iy, 
			Nz, index, output);


  /* last entry in index file */
  fprintf (cldfile, "%d %s %d %f %d\n", 
	   index, cldcolfilename, 
	   (sarray[sarrayc-1]).nc, 
	   (sarray[sarrayc-1]).tau,
	   keyc);

  output->ipaweight[index] = (double) keyc / (double) narray;

  fclose (cldfile);

  output->nipa=index+1;
  
  free(sarray);
  free(sarray1);
  
  return 0;
}





static int read_wc_ipa (input_struct input, output_struct *output)
{
  int ic=0, min_columns=0, max_columns=0, max_length=0, status=0;
  char ***string=NULL;
  char *dummy=NULL;
  

  if (!input.quiet)
    fprintf (stderr, " ... reading cloud data from %s\n", input.filename[FN_WCLOUD_IPA]);
      
  if ((status = ASCII_checkfile (input.filename[FN_WCLOUD_IPA], &(output->nipa), 
				 &min_columns, &max_columns, &max_length)) != 0) 
    return status;
  
  if (min_columns<2) {
    fprintf (stderr, "Error, need at least 2 columns in independent pixel file %s\n", 
	     input.filename[FN_WCLOUD_IPA]);
    return -1;
  }
  
  if ((status = ASCII_calloc_string(&string, output->nipa, max_columns, FILENAME_MAX+1)) != 0) 
    return status;
  
  if ((status = ASCII_readfile(input.filename[FN_WCLOUD_IPA], string)) != 0) 
    return status;
  

  /* allocate memory for IPA arrays */
  output->wcipa     = calloc (output->nipa, sizeof(cld_out_struct));
  output->ipaweight = calloc (output->nipa, sizeof(float));
  
  for (ic=0; ic<output->nipa; ic++) {
    output->ipaweight[ic] = (float) strtod (string[ic][1], &dummy);
    
    if (!input.quiet)
      fprintf (stderr, " ... reading %s\n", string[ic][0]);
    
    if (strlen(input.filename[FN_WCLOUD_PROP]) > 0) {
      status = read_cld_file (input, output,
			      string[ic][0], &(output->wcipa[ic]), 0);
      if (status!=0) {
	fprintf (stderr, "Error %d reading IPA %s\n", status, string[ic][0]);
	return status;
      }
    }
    else {
      status = read_and_convert_wc_file (input, output,
					 string[ic][0], &(output->wcipa[ic]));
      if (status!=0) {
	fprintf (stderr, "Error %d reading IPA %s\n", status, string[ic][0]);
	return status;
      }
    }
  }    
  
  /* Read optical properties file if available */
  if (strlen(input.filename[FN_WCLOUD_PROP]) > 0) {
    
    status = read_wcloud_prop (input, output);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading water cloud property file %s\n", 
	       status, input.filename[FN_WCLOUD_PROP]);
      return status;
    }

    if (input.verbose)
      fprintf (stderr, " ... read water cloud optical properties from %s\n", 
	       input.filename[FN_WCLOUD_PROP]);
  }


  if (input.verbose)
    fprintf (stderr, " ... read %d independent water cloud columns from %s\n", 
	     output->nipa, input.filename[FN_WCLOUD_IPA]);

  return 0;
}



static int read_ic_ipa (input_struct input, output_struct *output)
{
  int ic=0, min_columns=0, max_columns=0, max_length=0, status=0;
  char ***string=NULL;
  char *dummy=NULL;

  float *tmpweight=NULL;
  int tmpn=0;
  

  if (!input.quiet)
    fprintf (stderr, " ... reading cloud data from %s\n", input.filename[FN_ICLOUD_IPA]);
  
  if ((status = ASCII_checkfile (input.filename[FN_ICLOUD_IPA], &tmpn, 
				 &min_columns, &max_columns, &max_length)) != 0) 
     return status;
  
  if (min_columns<2) {
    fprintf (stderr, "Error, need at least 2 columns in independent pixel file %s\n", 
	     input.filename[FN_ICLOUD_IPA]);
    return -1;
  }
  

  /* check if water cloud and ice cloud files have same number of rows */
  if (input.wcipa && tmpn != output->nipa) {
    fprintf (stderr, "Error. %s and %s have different number of rows, %d and %d.\n",
	     input.filename[FN_ICLOUD_IPA], input.filename[FN_WCLOUD_IPA],
	     tmpn, output->nipa);
    return -1;
  }
  
  output->nipa = tmpn;
  

  /* read ASCII file */
  if ((status = ASCII_calloc_string(&string, output->nipa, max_columns, FILENAME_MAX+1)) != 0) 
    return status;
  
  if ((status = ASCII_readfile(input.filename[FN_ICLOUD_IPA], string)) != 0) 
    return status;
  

  /* allocate memory for IPA arrays */
  output->icipa     = calloc (output->nipa, sizeof(cld_out_struct));
  tmpweight = calloc (output->nipa, sizeof(float));
  

  /* extract weight */
  for (ic=0; ic<output->nipa; ic++)
    tmpweight[ic] = (float) strtod (string[ic][1], &dummy);

  if (input.wcipa) {
    for (ic=0; ic<output->nipa; ic++) {
      if (output->ipaweight[ic] != tmpweight[ic]) {
	fprintf (stderr, "Error, water and ice cloud weights are not identical:\n");
	fprintf (stderr, "       %f and %f in row %d\n", 
		 output->ipaweight[ic], tmpweight[ic], ic);
	return -1;
      }
    }
    free(tmpweight);
  }
  else 
    output->ipaweight = tmpweight;
    
	

  /* now finally read the ice cloud files */



  for (ic=0; ic<output->nipa; ic++) {
    if (!input.quiet)
      fprintf (stderr, " ... reading %s\n", string[ic][0]);
    

    if (strlen(input.filename[FN_ICLOUD_PROP]) > 0) {
      status = read_cld_file (input, output,
			      string[ic][0], &(output->icipa[ic]), 0);
      if (status!=0) {
	fprintf (stderr, "Error %d reading IPA %s\n", status, string[ic][0]);
	return status;
      }
    }
    else {
      status = read_and_convert_ic_file (input, output,
					  string[ic][0], &(output->icipa[ic]));
      if (status!=0) {
	fprintf (stderr, "Error %d reading IPA %s\n", status, string[ic][0]);
	return status;
      }
    }
  }


  if (strlen(input.filename[FN_ICLOUD_PROP]) > 0) {
    
    status = read_icloud_prop (input, output);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading ice cloud property file %s\n", 
		 status, input.filename[FN_ICLOUD_PROP]);
      return status;
    }

    if (input.verbose)
      fprintf (stderr, " ... read ice cloud optical properties from %s\n", 
	       input.filename[FN_ICLOUD_PROP]);
  }


  if (input.verbose)
    fprintf (stderr, " ... read %d independent ice cloud columns from %s\n", 
	     output->nipa, input.filename[FN_ICLOUD_IPA]);

  return 0;
}
